/*
 * Buxton
 *
 * Copyright (C) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License)
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <poll.h>
#include <time.h>
#include <pthread.h>

#include <glib.h>
#include <glib-unix.h>

#include "buxton2.h"
#include "buxton_error.h"

#include "common.h"
#include "log.h"
#include "serialize.h"
#include "proto.h"

#ifndef EXPORT
#  define EXPORT __attribute__((visibility("default")))
#endif

struct bxt_req {
	guint32 msgid;
	struct buxton_layer *layer;
	char *key;

	buxton_response_callback callback;
	buxton_list_callback list_cb;
	void *data;

	buxton_notify_callback notify;
	void *notify_data;
};

struct bxt_noti_cb {
	int ref_cnt;
	buxton_notify_callback callback;
	void *data;
};

struct bxt_noti {
	guint id;
	char *layer_key; /* layer + <tab>(0x09) + key */
	gboolean reg;
	GList *callbacks; /* data: bxt_noti_cb */
	pthread_mutex_t cbs_lock;
};

struct bxt_noti_res {
	int res;
	struct buxton_layer *layer;
	char *key;
	buxton_response_callback callback;
	void *data;
};

struct buxton_client {
	int fd;
	guint fd_id;

	buxton_status_callback st_callback;
	void *st_data;

	GHashTable *req_cbs; /* key: msgid, value: bxt_req */
	GHashTable *noti_cbs; /* key: keyname, value: bxt_noti */
	pthread_mutex_t lock;
};

static GList *clients; /* data: buxton_client */
static pthread_mutex_t clients_lock = PTHREAD_MUTEX_INITIALIZER;
static guint32 client_msgid;

static int value_create(enum buxton_key_type type, void *value,
						struct buxton_value **new_value)
{
	struct buxton_value *val;

	if (!value || !new_value) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	val = calloc(1, sizeof(*val));
	if (!val) {
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	switch (type) {
	case BUXTON_TYPE_STRING:
		val->value.s = *(char **)value;
		break;
	case BUXTON_TYPE_INT32:
		val->value.i = *(int32_t *)value;
		break;
	case BUXTON_TYPE_UINT32:
		val->value.u = *(uint32_t *)value;
		break;
	case BUXTON_TYPE_INT64:
		val->value.i64 = *(int64_t *)value;
		break;
	case BUXTON_TYPE_UINT64:
		val->value.u64 = *(uint64_t *)value;
		break;
	case BUXTON_TYPE_DOUBLE:
		val->value.d = *(double *)value;
		break;
	case BUXTON_TYPE_BOOLEAN:
		val->value.b = *(int32_t *)value;
		break;
	default:
		free(val);
		bxt_err("Unknow type %d", type);
		return BUXTON_ERROR_INVALID_PARAMETER;
	}
	val->type = type;
	*new_value = val;

	return BUXTON_ERROR_NONE;
}

EXPORT struct buxton_value *buxton_value_create_string(const char *s)
{
	struct buxton_value *v;
	char *str;
	int ret;

	if (!s) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return NULL;
	}

	str = strdup(s);
	if (!str) {
		bxt_err("out of memory");
		buxton_err_set_errno(BUXTON_ERROR_OUT_OF_MEMORY);
		return NULL;
	}

	ret = value_create(BUXTON_TYPE_STRING, &str, &v);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return NULL;
	}

	return v;
}

EXPORT struct buxton_value *buxton_value_create_int32(int32_t i)
{
	struct buxton_value *v;
	int ret;

	ret = value_create(BUXTON_TYPE_INT32, &i, &v);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return NULL;
	}
	return v;
}

EXPORT struct buxton_value *buxton_value_create_uint32(uint32_t u)
{
	struct buxton_value *v;
	int ret;

	ret = value_create(BUXTON_TYPE_UINT32, &u, &v);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return NULL;
	}
	return v;
}

EXPORT struct buxton_value *buxton_value_create_int64(int64_t i64)
{
	struct buxton_value *v;
	int ret;

	ret = value_create(BUXTON_TYPE_INT64, &i64, &v);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return NULL;
	}
	return v;
}

EXPORT struct buxton_value *buxton_value_create_uint64(uint64_t u64)
{
	struct buxton_value *v;
	int ret;

	ret = value_create(BUXTON_TYPE_UINT64, &u64, &v);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return NULL;
	}
	return v;
}

EXPORT struct buxton_value *buxton_value_create_double(double d)
{
	struct buxton_value *v;
	int ret;

	ret = value_create(BUXTON_TYPE_DOUBLE, &d, &v);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return NULL;
	}
	return v;
}

EXPORT struct buxton_value *buxton_value_create_boolean(int32_t b)
{
	struct buxton_value *v;
	int ret;

	ret = value_create(BUXTON_TYPE_BOOLEAN, &b, &v);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return NULL;
	}
	return v;
}

EXPORT int buxton_value_get_type(const struct buxton_value *val,
		enum buxton_key_type *type)
{
	if (!val || !type) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	*type = val->type;

	return 0;
}

static int value_get(const struct buxton_value *val, void *dest,
		enum buxton_key_type type)
{
	if (!val || !dest) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	if (val->type != type) {
		bxt_err("Invalid parameter : type mismatch %d %d", val->type, type);
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	switch (type) {
	case BUXTON_TYPE_STRING:
		*(char **)dest = val->value.s;
		break;
	case BUXTON_TYPE_INT32:
		*(int32_t *)dest = val->value.i;
		break;
	case BUXTON_TYPE_UINT32:
		*(uint32_t *)dest = val->value.u;
		break;
	case BUXTON_TYPE_INT64:
		*(int64_t *)dest = val->value.i64;
		break;
	case BUXTON_TYPE_UINT64:
		*(uint64_t *)dest = val->value.u64;
		break;
	case BUXTON_TYPE_DOUBLE:
		*(double *)dest = val->value.d;
		break;
	case BUXTON_TYPE_BOOLEAN:
		*(int32_t *)dest = val->value.b;
		break;
	default:
		bxt_err("Unknow type %d", type);
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_value_get_string(const struct buxton_value *val,
		const char **s)
{
	int ret;

	ret = value_get(val, s, BUXTON_TYPE_STRING);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return -1;
	}
	return 0;
}

EXPORT int buxton_value_get_int32(const struct buxton_value *val, int32_t *i)
{
	int ret;

	ret = value_get(val, i, BUXTON_TYPE_INT32);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return -1;
	}
	return 0;
}

EXPORT int buxton_value_get_uint32(const struct buxton_value *val, uint32_t *u)
{
	int ret;

	ret = value_get(val, u, BUXTON_TYPE_UINT32);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return -1;
	}
	return 0;
}

EXPORT int buxton_value_get_int64(const struct buxton_value *val, int64_t *i64)
{
	int ret;

	ret = value_get(val, i64, BUXTON_TYPE_INT64);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return -1;
	}
	return 0;
}

EXPORT int buxton_value_get_uint64(const struct buxton_value *val,
		uint64_t *u64)
{
	int ret;

	ret = value_get(val, u64, BUXTON_TYPE_UINT64);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return -1;
	}
	return 0;
}

EXPORT int buxton_value_get_double(const struct buxton_value *val, double *d)
{
	int ret;

	ret = value_get(val, d, BUXTON_TYPE_DOUBLE);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return -1;
	}
	return 0;
}

EXPORT int buxton_value_get_boolean(const struct buxton_value *val, int32_t *b)
{
	int ret;

	ret = value_get(val, b, BUXTON_TYPE_BOOLEAN);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return -1;
	}
	return 0;
}

EXPORT struct buxton_value *buxton_value_duplicate(
		const struct buxton_value *val)
{
	struct buxton_value *_val;

	if (!val) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return NULL;
	}

	_val = malloc(sizeof(*_val));
	if (!_val) {
		bxt_err("out of memory");
		buxton_err_set_errno(BUXTON_ERROR_OUT_OF_MEMORY);
		return NULL;
	}

	*_val = *val;

	if (val->type == BUXTON_TYPE_STRING && val->value.s) {
		_val->value.s = strdup(val->value.s);
		if (!_val->value.s) {
			free(_val);
			bxt_err("out of memory");
			buxton_err_set_errno(BUXTON_ERROR_OUT_OF_MEMORY);
			return NULL;
		}
	}

	return _val;
}

EXPORT void buxton_value_free(struct buxton_value *val)
{
	value_free(val);
	free(val);
}

EXPORT struct buxton_layer *buxton_create_layer(const char *layer_name)
{
	struct buxton_layer *layer;
	int ret;

	ret = layer_create(layer_name, &layer);
	if (ret != BUXTON_ERROR_NONE) {
		buxton_err_set_errno(ret);
		return NULL;
	}

	return layer;
}

EXPORT const char *buxton_layer_get_name(const struct buxton_layer *layer)
{
	if (!layer) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return NULL;
	}

	return layer->name;
}

EXPORT void buxton_layer_set_uid(struct buxton_layer *layer, uid_t uid)
{
	if (!layer) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return;
	}

	layer->uid = uid;
}

EXPORT void buxton_layer_set_type(struct buxton_layer *layer,
		enum buxton_layer_type type)
{
	if (!layer) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return;
	}
	switch (type) {
	case BUXTON_LAYER_NORMAL:
	case BUXTON_LAYER_BASE:
		break;
	default:
		bxt_err("Unknow type %d", type);
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return;
	}

	layer->type = type;
}

EXPORT void buxton_free_layer(struct buxton_layer *layer)
{
	layer_unref(layer);
}

static int create_req(const struct buxton_layer *layer,
		const char *key, buxton_response_callback callback,
		buxton_list_callback list_cb, void *data, struct bxt_req **req)
{
	struct bxt_req *_req;

	if (!layer || !(callback || list_cb)) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	_req = calloc(1, sizeof(*_req));
	if (!_req) {
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	if (key) {
		_req->key = strdup(key);
		if (!_req->key) {
			free(_req);
			bxt_err("out of memory");
			return BUXTON_ERROR_OUT_OF_MEMORY;
		}
	}

	_req->layer = layer_ref((struct buxton_layer *)layer);
	_req->callback = callback;
	_req->list_cb = list_cb;
	_req->data = data;
	_req->msgid = __atomic_fetch_add(&client_msgid, 1, __ATOMIC_RELAXED);

	*req = _req;

	return BUXTON_ERROR_NONE;
}

static int find_noti(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		struct bxt_noti **noti)
{
	char *lykey;
	struct bxt_noti *_noti;
	int ret;

	if (!client || !layer || !key || !*key || !noti) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	ret = get_search_key(layer, key, NULL, &lykey);
	if (ret != BUXTON_ERROR_NONE)
		return ret;

	pthread_mutex_lock(&client->lock);
	_noti = g_hash_table_lookup(client->noti_cbs, lykey);

	free(lykey);

	*noti = _noti;

	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

static GList *copy_noti_callbacks(GList *callbacks)
{
	GList *l;
	GList *copy;

	if (!callbacks) {
		bxt_err("Invalid parameter");
		return NULL;
	}

	for (l = callbacks; l; l = g_list_next(l)) {
		struct bxt_noti_cb *noticb = l->data;
		noticb->ref_cnt++;
	}
	copy = g_list_copy(callbacks);

	return copy;
}

static GList *free_copy_callbacks(GList *callbacks, GList *copy_callbacks)
{
	GList *l;
	GList *ll;

	g_list_free(copy_callbacks);

	for (l = callbacks, ll = g_list_next(l); l;
			l = ll, ll = g_list_next(ll)) {
		struct bxt_noti_cb *noticb = l->data;

		noticb->ref_cnt--;
		if (noticb->ref_cnt == 0) {
			callbacks = g_list_delete_link(callbacks, l);
			free(noticb);
		}
	}

	return callbacks;
}

static int proc_msg_noti(struct buxton_client *client, uint8_t *data, int len)
{
	int r;
	struct request rqst;
	struct bxt_noti *noti;
	GList *l;
	GList *copy_callbacks;

	if (!client || !data || len <= 0) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = deserialz_request(data, len, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("proc noti: deserialize errno %d", r);
		return r;
	}

	noti = NULL;
	r = find_noti(client, rqst.layer, rqst.key, &noti);
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("proc noti: '%s' '%s' not registered (%d)",
				rqst.layer->name, rqst.key, r);
		free_request(&rqst);
		return r;
	}

	if (!noti) {
		bxt_err("proc noti: '%s' '%s' callback not exist",
				rqst.layer->name, rqst.key);
		free_request(&rqst);
		return BUXTON_ERROR_NOT_EXIST;
	}

	pthread_mutex_lock(&noti->cbs_lock);
	copy_callbacks = copy_noti_callbacks(noti->callbacks);

	for (l = copy_callbacks; l; l = g_list_next(l)) {
		struct bxt_noti_cb *noticb = l->data;
		if (noticb->callback)
			noticb->callback(rqst.layer, rqst.key, rqst.val, noticb->data);
	}

	noti->callbacks = free_copy_callbacks(noti->callbacks, copy_callbacks);
	pthread_mutex_unlock(&noti->cbs_lock);
	free_request(&rqst);

	return BUXTON_ERROR_NONE;
}

static int add_noti(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		struct bxt_noti **noti)
{
	char *lykey;
	struct bxt_noti *_noti;
	int ret;

	if (!client || !layer || !key || !*key || !noti) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	ret = get_search_key(layer, key, NULL, &lykey);
	if (ret != BUXTON_ERROR_NONE)
		return ret;

	pthread_mutex_lock(&client->lock);
	_noti = g_hash_table_lookup(client->noti_cbs, lykey);
	if (_noti) {
		free(lykey);
		*noti = _noti;
		pthread_mutex_unlock(&client->lock);
		return BUXTON_ERROR_NONE;
	}

	_noti = calloc(1, sizeof(*_noti));
	if (!_noti) {
		free(lykey);
		pthread_mutex_unlock(&client->lock);
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}
	_noti->layer_key = lykey;
	pthread_mutex_init(&_noti->cbs_lock, NULL);
	g_hash_table_insert(client->noti_cbs, lykey, _noti);

	*noti = _noti;

	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

static int add_noticb(struct bxt_noti *noti, buxton_notify_callback notify,
		void *notify_data)
{
	GList *l;
	struct bxt_noti_cb *noticb;

	if (!noti || !notify) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	pthread_mutex_lock(&noti->cbs_lock);
	for (l = noti->callbacks; l; l = g_list_next(l)) {
		noticb = l->data;

		if (noticb->callback == notify) {
			pthread_mutex_unlock(&noti->cbs_lock);
			return BUXTON_ERROR_EXIST;
		}
	}

	noticb = calloc(1, sizeof(*noticb));
	if (!noticb) {
		pthread_mutex_unlock(&noti->cbs_lock);
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	noticb->ref_cnt = 1;
	noticb->callback = notify;
	noticb->data = notify_data;

	noti->callbacks = g_list_append(noti->callbacks, noticb);

	pthread_mutex_unlock(&noti->cbs_lock);
	return BUXTON_ERROR_NONE;
}

static void del_noti(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key)
{
	char *lykey;
	int ret;

	if (!client || !layer || !key || !*key) {
		bxt_err("Invalid parameter");
		return;
	}

	ret = get_search_key(layer, key, NULL, &lykey);
	if (ret != BUXTON_ERROR_NONE)
		return;

	pthread_mutex_lock(&client->lock);
	g_hash_table_remove(client->noti_cbs, lykey);
	pthread_mutex_unlock(&client->lock);

	free(lykey);
}

static void proc_msg_notify(struct buxton_client *client, struct bxt_req *req,
		struct response *resp)
{
	struct bxt_noti *noti;
	int r;

	if (!client || !req || !resp) {
		bxt_err("Invalid parameter");
		return;
	}

	if (resp->res == BUXTON_ERROR_NONE) {
		r = add_noti(client, req->layer, req->key, &noti);
		if (r != BUXTON_ERROR_NONE) {
			resp->res = r;
			bxt_err("add noti: errno %d", r);
		} else {
			if (noti->reg == FALSE)
				noti->reg = (resp->res == 0);

			r = add_noticb(noti, req->notify, req->notify_data);
			if (r != BUXTON_ERROR_NONE && r != BUXTON_ERROR_EXIST) {
				resp->res = r;
				bxt_err("add noticb: errno %d", r);
			}
		}
	}

	if (req->callback)
		req->callback(resp->res, req->layer, req->key, resp->val, req->data);
	else
		bxt_err("proc msg(%d,%d): callback not exist", resp->msgid, resp->type);
}

static int proc_msg_res(struct buxton_client *client, uint8_t *data, int len)
{
	int r;
	struct response resp;
	struct bxt_req *req;

	if (!client || !data || len <= 0) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = deserialz_response(data, len, &resp);
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("proc msg: deserialize errno %d", r);
		return r;
	}

	pthread_mutex_lock(&client->lock);
	req = g_hash_table_lookup(client->req_cbs,
			GUINT_TO_POINTER(resp.msgid));
	pthread_mutex_unlock(&client->lock);
	if (!req) {
		bxt_err("proc msg: msgid %d not exist", resp.msgid);
		free_response(&resp);
		return BUXTON_ERROR_NONE;
	}

	switch (resp.type) {
	case MSG_LIST:
		if (req->list_cb)
			req->list_cb(resp.res, req->layer, resp.names, resp.nmlen, req->data);
		else
			bxt_err("proc msg(%d,%d): list_cb not exist", resp.msgid, resp.type);

		break;
	case MSG_NOTIFY:
		proc_msg_notify(client, req, &resp);
		break;
	case MSG_UNNOTIFY:
		del_noti(client, req->layer, req->key);

		if (req->callback)
			req->callback(resp.res, req->layer, req->key, resp.val, req->data);
		else
			bxt_err("proc msg(%d,%d): callback not exist", resp.msgid, resp.type);

		break;
	default:
		if (req->callback)
			req->callback(resp.res, req->layer, req->key, resp.val,
				req->data);
		else
			bxt_err("proc msg(%d,%d): callback not exist", resp.msgid, resp.type);

		break;
	}

	free_response(&resp);

	pthread_mutex_lock(&client->lock);
	g_hash_table_remove(client->req_cbs, GUINT_TO_POINTER(resp.msgid));
	pthread_mutex_unlock(&client->lock);

	return BUXTON_ERROR_NONE;
}

static void proc_msg_cb(void *user_data,
		enum message_type type, uint8_t *data, int32_t len)
{
	struct buxton_client *client = user_data;

	switch (type) {
	case MSG_NOTI:
		proc_msg_noti(client, data, len);
		break;
	case MSG_SET:
	case MSG_GET:
	case MSG_CREAT:
	case MSG_UNSET:
	case MSG_LIST:
	case MSG_NOTIFY:
	case MSG_UNNOTIFY:
	case MSG_SET_WP:
	case MSG_SET_RP:
	case MSG_GET_WP:
	case MSG_GET_RP:
	case MSG_CTRL:
		proc_msg_res(client, data, len);
		break;
	default:
		bxt_err("proc msg: unknown message type %d", type);
		break;
	}
}

static int proc_msg(struct buxton_client *client)
{
	int r;
	GList *f;

	f = g_list_find(clients, client);
	if (!f) {
		bxt_dbg("recv msg: cli %p removed\n", client);
		return BUXTON_ERROR_NONE;
	}

	r = proto_recv_async(client->fd, proc_msg_cb, client);
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("recv msg: fd %d errno %d", client->fd, r);
		return r;
	}

	return BUXTON_ERROR_NONE;
}

#define TS_SUB(a, b) (((a)->tv_sec - (b)->tv_sec) * 1000 \
		+ ((a)->tv_nsec - (b)->tv_nsec) / 1000000)
static int wait_msg(struct buxton_client *client, guint32 msgid)
{
	int r, ret;
	struct pollfd fds[1];
	struct timespec to;
	struct timespec t;
	int32_t ms;
	struct bxt_req *req;

	if (client->fd < 0) {
		bxt_err("Invalid parameter (%d)", client->fd);
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	fds[0].fd = client->fd;
	fds[0].events = POLLIN;
	fds[0].revents = 0;

	clock_gettime(CLOCK_MONOTONIC, &t);
	to.tv_sec = t.tv_sec + 5; /* TIMEOUT 5 seconds */
	to.tv_nsec = t.tv_nsec;

	ms = TS_SUB(&to, &t);

	while (ms > 0) {
		r = poll(fds, 1, ms > 10 ? 10 : ms);
		ret = BUXTON_ERROR_NONE;

		switch (r) {
		case -1:
			if (errno == EINTR) {
				clock_gettime(CLOCK_MONOTONIC, &t);
				ms = TS_SUB(&to, &t);
				continue;
			}
			bxt_err("wait response: poll: fd %d errno %d",
					client->fd, errno);
			ret = BUXTON_ERROR_IO_ERROR;
			break;
		case 0:
			/* poll timeout, do nothing */
			break;
		default:
			ret = proc_msg(client);
			break;
		}

		/* poll or proc error */
		if (ret != BUXTON_ERROR_NONE) {
			pthread_mutex_lock(&client->lock);
			g_hash_table_remove(client->req_cbs,
					GUINT_TO_POINTER(msgid));
			pthread_mutex_unlock(&client->lock);

			return ret;
		}

		pthread_mutex_lock(&client->lock);
		req = g_hash_table_lookup(client->req_cbs,
				GUINT_TO_POINTER(msgid));
		/* req is processed */
		if (!req) {
			pthread_mutex_unlock(&client->lock);
			return BUXTON_ERROR_NONE;
		}

		pthread_mutex_unlock(&client->lock);
		clock_gettime(CLOCK_MONOTONIC, &t);
		ms = TS_SUB(&to, &t);
	}

	bxt_err("wait response: timeout");

	pthread_mutex_lock(&client->lock);
	g_hash_table_remove(client->req_cbs, GUINT_TO_POINTER(msgid));
	pthread_mutex_unlock(&client->lock);

	return BUXTON_ERROR_TIME_OUT;
}

static void free_req(struct bxt_req *req)
{
	if (!req)
		return;

	layer_unref(req->layer);
	free(req->key);
	free(req);
}

static int send_req(struct buxton_client *client, const struct request *rqst)
{
	int r;
	uint8_t *data;
	int len;

	if (client->fd < 0) {
		bxt_err("Invalid parameter(%d)", client->fd);
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = serialz_request(rqst, &data, &len);
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("send req: serialize errno %d", r);
		return r;
	}

	r = proto_send(client->fd, rqst->type, data, len);
	if (r != BUXTON_ERROR_NONE)
		bxt_err("send req: errno %d", r);

	free(data);

	return r;
}

static int set_value(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		const struct buxton_value *val,
		buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;

	if (!client || !layer || !key || !*key || !val || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = create_req(layer, key, callback, NULL, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_SET;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = (char *)key;
	rqst.val = (struct buxton_value *)val;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_set_value(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		const struct buxton_value *val,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int ret;

	pthread_mutex_lock(&clients_lock);
	ret = set_value(client, layer, key, val, callback, user_data, &req);
	if (ret != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(ret);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void set_value_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}
	resp->res = status;
}

EXPORT int buxton_set_value_sync(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		const struct buxton_value *val)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = set_value(client, layer, key, val, set_value_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static int get_value(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;

	if (!client || !layer || !key || !*key || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = create_req(layer, key, callback, NULL, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_GET;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = (char *)key;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_get_value(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int ret;
	pthread_mutex_lock(&clients_lock);
	ret = get_value(client, layer, key, callback, user_data, &req);
	if (ret != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(ret);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void get_value_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;

	if (!status)
		resp->val = buxton_value_duplicate(val);
}

EXPORT int buxton_get_value_sync(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		struct buxton_value **val)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	if (!val) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = get_value(client, layer, key, get_value_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	*val = resp.val;

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static int list_keys(struct buxton_client *client,
		const struct buxton_layer *layer,
		buxton_list_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;

	if (!client || !layer || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = create_req(layer, NULL, NULL, callback, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_LIST;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);

	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_list_keys(struct buxton_client *client,
		const struct buxton_layer *layer,
		buxton_list_callback callback, void *user_data)
{
	struct bxt_req *req;
	int ret;
	pthread_mutex_lock(&clients_lock);
	ret = list_keys(client, layer, callback, user_data, &req);
	if (ret != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(ret);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void list_keys_sync_cb(int status, const struct buxton_layer *layer,
		char * const *names, unsigned int len, void *user_data)
{
	struct response *resp = user_data;
	char **nms;
	char **_names;
	int i;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;

	if (resp->res != BUXTON_ERROR_NONE)
		return;

	nms = calloc(len + 1, sizeof(void *));
	if (!nms) {
		bxt_err("out of memory");
		resp->res = BUXTON_ERROR_OUT_OF_MEMORY;
		return;
	}

	/* steal allocated names */
	_names = (char **)names;
	for (i = 0; i < len; i++) {
		nms[i] = _names[i];
		_names[i] = NULL;
	}
	nms[i] = NULL; /* NULL-terminated */

	resp->names = nms;
	resp->nmlen = len;
}

EXPORT int buxton_list_keys_sync(struct buxton_client *client,
		const struct buxton_layer *layer,
		char ***names, unsigned int *len)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	if (!names) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = list_keys(client, layer, list_keys_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	*names = resp.names;

	if (len)
		*len = resp.nmlen;

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static gboolean call_resp_cb(gpointer data)
{
	struct bxt_noti_res *res = data;

	if (!res || !res->callback) {
		bxt_err("Invalid parameter");
		return G_SOURCE_REMOVE;
	}

	buxton_err_set_errno(res->res);
	res->callback(buxton_err_get_errno(), res->layer, res->key, NULL, res->data);

	layer_unref(res->layer);
	free(res->key);
	free(res);

	return G_SOURCE_REMOVE;
}

static int call_resp(int status, const struct buxton_layer *layer,
		const char *key, buxton_response_callback callback,
		void *user_data)
{
	struct bxt_noti_res *res;

	if (!layer || !key || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	res = calloc(1, sizeof(*res));
	if (!res) {
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}
	res->key = strdup(key);
	if (!res->key) {
		free(res);
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	res->layer = layer_ref((struct buxton_layer *)layer);
	res->res = status;
	res->callback = callback;
	res->data = user_data;

	g_idle_add(call_resp_cb, res);

	return BUXTON_ERROR_NONE;
}

static int register_noti(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_notify_callback notify, void *notify_data,
		buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;

	r = create_req(layer, key, callback, NULL, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	req->notify = notify;
	req->notify_data = notify_data;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_NOTIFY;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = (char *)key;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);

	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_register_notification(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_notify_callback notify, void *notify_data,
		buxton_response_callback callback, void *user_data)
{
	int r;
	struct bxt_noti *noti;
	struct bxt_req *req;

	if (!client || !layer || !key || !*key || !notify || !callback) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	pthread_mutex_lock(&clients_lock);
	r = find_noti(client, layer, key, &noti);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (noti && noti->reg == TRUE) {
		r = add_noticb(noti, notify, notify_data);
		pthread_mutex_unlock(&clients_lock);
		r = call_resp(r, layer, key, callback, user_data);
		if (r != BUXTON_ERROR_NONE) {
			buxton_err_set_errno(r);
			return -1;
		}
		return 0;
	}

	r = register_noti(client, layer, key, notify, notify_data, callback,
			user_data, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void reg_noti_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;
}

EXPORT int buxton_register_notification_sync(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_notify_callback notify, void *notify_data)
{
	int r;
	struct bxt_noti *noti;
	struct bxt_req *req;
	struct response resp;

	if (!client || !layer || !key || !*key || !notify) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	pthread_mutex_lock(&clients_lock);
	r = find_noti(client, layer, key, &noti);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (noti && noti->reg == TRUE) {
		pthread_mutex_unlock(&clients_lock);
		r = add_noticb(noti, notify, notify_data);
		if (r != BUXTON_ERROR_NONE) {
			buxton_err_set_errno(r);
			return -1;
		}
		return 0;
	}

	memset(&resp, 0, sizeof(resp));

	r = register_noti(client, layer, key, notify, notify_data,
			reg_noti_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static int del_noticb(struct bxt_noti *noti, buxton_notify_callback notify,
		int *count)
{
	GList *l;
	GList *ll;
	gboolean f;
	int cnt;

	if (!noti || !notify) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	cnt = 0;
	f = FALSE;

	pthread_mutex_lock(&noti->cbs_lock);
	for (l = noti->callbacks, ll = g_list_next(l); l;
			l = ll, ll = g_list_next(ll)) {
		struct bxt_noti_cb *noticb = l->data;

		if (noticb->callback == notify) {
			f = TRUE;
			noticb->ref_cnt--;
			if (noticb->ref_cnt == 0) {
				noti->callbacks = g_list_delete_link(noti->callbacks, l);
				free(noticb);
			}
		} else {
			cnt++;
		}
	}
	pthread_mutex_unlock(&noti->cbs_lock);

	if (!f) {
		bxt_err("no such a notify");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	if (count)
		*count = cnt;

	return BUXTON_ERROR_NONE;
}

static int unregister_noti(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;

	r = create_req(layer, key, callback, NULL, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_UNNOTIFY;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = (char *)key;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_unregister_notification(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_notify_callback notify,
		buxton_response_callback callback, void *user_data)
{
	int r;
	struct bxt_noti *noti;
	struct bxt_req *req;
	int cnt;

	if (!client || !layer || !key || !*key || !notify || !callback) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	pthread_mutex_lock(&clients_lock);
	r = find_noti(client, layer, key, &noti);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (!noti) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(BUXTON_ERROR_NOT_EXIST);
		return -1;
	}

	r = del_noticb(noti, notify, &cnt);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		r = call_resp(r, layer, key, callback, user_data);
		if (r != BUXTON_ERROR_NONE) {
			buxton_err_set_errno(r);
			return -1;
		}
		return 0;
	}

	if (cnt || noti->reg == FALSE) {
		pthread_mutex_unlock(&clients_lock);
		r = call_resp(BUXTON_ERROR_NONE, layer, key, callback, user_data);
		if (r != BUXTON_ERROR_NONE) {
			buxton_err_set_errno(r);
			return -1;
		}
		return 0;
	}

	r = unregister_noti(client, layer, key, callback, user_data, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void unreg_noti_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;
}

EXPORT int buxton_unregister_notification_sync(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_notify_callback notify)
{
	int r;
	struct bxt_noti *noti;
	struct bxt_req *req;
	int cnt;
	struct response resp;

	if (!client || !layer || !key || !*key || !notify) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	pthread_mutex_lock(&clients_lock);
	r = find_noti(client, layer, key, &noti);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (!noti) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(BUXTON_ERROR_NOT_EXIST);
		return -1;
	}

	r = del_noticb(noti, notify, &cnt);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (cnt || noti->reg == FALSE) {
		pthread_mutex_unlock(&clients_lock);
		return 0;
	}

	memset(&resp, 0, sizeof(resp));

	r = unregister_noti(client, layer, key, unreg_noti_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static int create_value(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		const char *read_privilege, const char *write_privilege,
		const struct buxton_value *val,
		buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;

	if (!client || !layer || !key || !*key || !read_privilege
			|| !write_privilege || !val || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;

	}

	r = create_req(layer, key, callback, NULL, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_CREAT;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = (char *)key;
	rqst.rpriv = (char *)read_privilege;
	rqst.wpriv = (char *)write_privilege;
	rqst.val = (struct buxton_value *)val;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_create_value(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		const char *read_privilege, const char *write_privilege,
		const struct buxton_value *val,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int ret;

	pthread_mutex_lock(&clients_lock);
	ret = create_value(client, layer, key, read_privilege, write_privilege,
			val, callback, user_data, &req);
	if (ret != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(ret);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void create_value_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;
}

EXPORT int buxton_create_value_sync(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		const char *read_privilege, const char *write_privilege,
		const struct buxton_value *val)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = create_value(client, layer, key, read_privilege, write_privilege,
			val, create_value_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static int unset_value(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;

	if (!client || !layer || !key || !*key || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = create_req(layer, key, callback, NULL, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_UNSET;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = (char *)key;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_unset_value(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int ret;

	pthread_mutex_lock(&clients_lock);
	ret = unset_value(client, layer, key, callback, user_data, &req);
	if (ret != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(ret);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void unset_value_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;
}

EXPORT int buxton_unset_value_sync(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = unset_value(client, layer, key, unset_value_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static int set_priv(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		enum buxton_priv_type type,
		const char *privilege,
		buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;
	struct buxton_value val;

	if (!client || !layer || !key || !*key || !privilege || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	if (type <= BUXTON_PRIV_UNKNOWN || type >= BUXTON_PRIV_MAX) {
		bxt_err("Invalid parameter (%d)", type);
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = create_req(layer, key, callback, NULL, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = type == BUXTON_PRIV_READ ? MSG_SET_RP : MSG_SET_WP;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = (char *)key;
	rqst.val = &val;

	val.type = BUXTON_TYPE_PRIVILEGE;
	val.value.s = (char *)privilege;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_set_privilege(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		enum buxton_priv_type type,
		const char *privilege,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int r;

	pthread_mutex_lock(&clients_lock);
	r = set_priv(client, layer, key, type, privilege,
			callback, user_data, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void set_priv_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;
}

EXPORT int buxton_set_privilege_sync(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		enum buxton_priv_type type,
		const char *privilege)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = set_priv(client, layer, key, type, privilege,
			set_priv_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static int get_priv(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		enum buxton_priv_type type,
		buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;

	if (!client || !layer || !key || !*key || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	if (type <= BUXTON_PRIV_UNKNOWN || type >= BUXTON_PRIV_MAX) {
		bxt_err("Invalid parameter (%d)", type);
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = create_req(layer, key, callback, NULL, user_data, &req);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = type == BUXTON_PRIV_READ ? MSG_GET_RP : MSG_GET_WP;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = (char *)key;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);
	if (r  != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_get_privilege(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		enum buxton_priv_type type,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int r;

	pthread_mutex_lock(&clients_lock);
	r = get_priv(client, layer, key, type, callback, user_data, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void get_priv_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;

	if (status == BUXTON_ERROR_NONE)
		resp->val = buxton_value_duplicate(val);
}

EXPORT int buxton_get_privilege_sync(struct buxton_client *client,
		const struct buxton_layer *layer, const char *key,
		enum buxton_priv_type type,
		char **privilege)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	if (!privilege) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = get_priv(client, layer, key, type, get_priv_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	*privilege = resp.val->value.s;
	resp.val->value.s = NULL;
	buxton_value_free(resp.val);

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void security_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;
}

static int security_control(struct buxton_client *client,
		int32_t enable, buxton_response_callback callback, void *user_data,
		struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;
	struct buxton_layer *layer;
	struct buxton_value val;

	if (!client || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = layer_create("dummy", &layer);
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = create_req(layer, NULL, callback, NULL, user_data, &req);
	layer_unref(layer);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_CTRL;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = strdup("set_security_mode");
	rqst.val = &val;

	val.type = BUXTON_TYPE_BOOLEAN;
	val.value.b = enable;

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);

	free(rqst.key);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_enable_security(struct buxton_client *client,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int r;

	pthread_mutex_lock(&clients_lock);
	r = security_control(client, TRUE, callback, user_data, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

EXPORT int buxton_enable_security_sync(struct buxton_client *client)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = security_control(client, TRUE, security_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

EXPORT int buxton_disable_security(struct buxton_client *client,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int r;

	pthread_mutex_lock(&clients_lock);
	r = security_control(client, FALSE, callback, user_data, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

EXPORT int buxton_disable_security_sync(struct buxton_client *client)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = security_control(client, FALSE, security_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void update_label_sync_cb(int status, const struct buxton_layer *layer,
		const char *key, const struct buxton_value *val,
		void *user_data)
{
	struct response *resp = user_data;

	if (!resp) {
		bxt_err("Invalid parameter");
		return;
	}

	resp->res = status;
}

static int update_client_label(struct buxton_client *client,
		buxton_response_callback callback,
		void *user_data, struct bxt_req **new_req)
{
	int r;
	struct bxt_req *req;
	struct request rqst;
	struct buxton_layer *layer;

	if (!client || !callback) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = layer_create("dummy", &layer);
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = create_req(layer, NULL, callback, NULL, user_data, &req);
	layer_unref(layer);
	if (r != BUXTON_ERROR_NONE)
		return r;

	memset(&rqst, 0, sizeof(rqst));
	rqst.type = MSG_CTRL;
	rqst.msgid = req->msgid;
	rqst.layer = req->layer;
	rqst.key = strdup("update_client_label");

	pthread_mutex_lock(&client->lock);
	g_hash_table_insert(client->req_cbs, GUINT_TO_POINTER(req->msgid), req);

	r = send_req(client, &rqst);

	free(rqst.key);
	if (r != BUXTON_ERROR_NONE) {
		g_hash_table_remove(client->req_cbs,
				GUINT_TO_POINTER(req->msgid));
		pthread_mutex_unlock(&client->lock);
		return r;
	}

	*new_req = req;
	pthread_mutex_unlock(&client->lock);
	return BUXTON_ERROR_NONE;
}

EXPORT int buxton_update_client_label(struct buxton_client *client,
		buxton_response_callback callback, void *user_data)
{
	struct bxt_req *req;
	int r;

	pthread_mutex_lock(&clients_lock);
	r = update_client_label(client, callback, user_data, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

EXPORT int buxton_update_client_label_sync(struct buxton_client *client)
{
	int r;
	struct bxt_req *req;
	struct response resp;

	memset(&resp, 0, sizeof(resp));

	pthread_mutex_lock(&clients_lock);
	r = update_client_label(client, update_label_sync_cb, &resp, &req);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	r = wait_msg(client, req->msgid);
	if (r != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (resp.res != BUXTON_ERROR_NONE) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(resp.res);
		return -1;
	}

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

static void free_noti(struct bxt_noti *noti)
{
	if (!noti)
		return;

	pthread_mutex_lock(&noti->cbs_lock);
	g_list_free_full(noti->callbacks, (GDestroyNotify)free);
	pthread_mutex_unlock(&noti->cbs_lock);

	if (noti->id) {
		g_source_remove(noti->id);
		noti->id = 0;
	}

	free(noti->layer_key);
	free(noti);
}

static gboolean close_conn(gpointer data)
{
	struct buxton_client *cli = data;
	GList *f;
	buxton_status_callback callback = NULL;
	void * user_data;

	if (!cli)
		return G_SOURCE_REMOVE;

	pthread_mutex_lock(&clients_lock);
	f = g_list_find(clients, cli);
	if (!f) {
		bxt_dbg("%s: cli %p removed\n", __func__, cli);
		pthread_mutex_unlock(&clients_lock);
		return G_SOURCE_REMOVE;
	}

	if (cli->fd == -1) {
		pthread_mutex_unlock(&clients_lock);
		return G_SOURCE_REMOVE;
	}

	if (cli->fd_id) {
		g_source_remove(cli->fd_id);
		cli->fd_id = 0;
	}

	close(cli->fd);
	cli->fd = -1;

	if (cli->st_callback) {
		callback = cli->st_callback;
		user_data = cli->st_data;
	}
	pthread_mutex_unlock(&clients_lock);

	if (callback)
		callback(BUXTON_STATUS_CONNECTION_CLOSED, user_data);

	return G_SOURCE_REMOVE;
}

static void free_client(struct buxton_client *cli)
{
	GList *f;

	if (!cli)
		return;

	f = g_list_find(clients, cli);
	if (f) {
		if (cli->fd_id) {
			g_source_remove(cli->fd_id);
			cli->fd_id = 0;
		}

		if (cli->fd != -1) {
			close(cli->fd);
			cli->fd = -1;
		}

		if (cli->st_callback)
			cli->st_callback(BUXTON_STATUS_DISCONNECTED, cli->st_data);
	}

	clients = g_list_remove(clients, cli);

	pthread_mutex_lock(&cli->lock);
	if (cli->req_cbs)
		g_hash_table_destroy(cli->req_cbs);

	if (cli->noti_cbs)
		g_hash_table_destroy(cli->noti_cbs);
	pthread_mutex_unlock(&cli->lock);

	free(cli);
}

int connect_server(const char *addr, int *new_fd)
{
	int fd;
	struct sockaddr_un sa;
	int r;

	if (!addr) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_NONBLOCK | SOCK_CLOEXEC, 0);
	if (fd == -1) {
		bxt_err("connect: socket errno %d", errno);
		return BUXTON_ERROR_IO_ERROR;
	}

	sa.sun_family = AF_UNIX;
	strncpy(sa.sun_path, addr, sizeof(sa.sun_path));
	sa.sun_path[sizeof(sa.sun_path) - 1] = '\0';

	r = connect(fd, (struct sockaddr *)&sa, sizeof(sa));
	if (r == -1) {
		if (errno == ENOENT)
			bxt_dbg("connect: '%s' not exist", addr);
		else
			bxt_dbg("connect: connect errno %d", errno);
		close(fd);
		return BUXTON_ERROR_IO_ERROR;
	}

	*new_fd = fd;

	return BUXTON_ERROR_NONE;
}

static gboolean recv_cb(gint fd, GIOCondition cond, gpointer data)
{
	struct buxton_client *cli = data;
	int r;
	GList *f;

	if (!cli) {
		bxt_err("Invalid parameter");
		return G_SOURCE_REMOVE;
	}

	pthread_mutex_lock(&clients_lock);
	f = g_list_find(clients, cli);
	if (!f) {
		bxt_dbg("recv %d: cli %p removed\n", fd, cli);
		pthread_mutex_unlock(&clients_lock);
		return G_SOURCE_REMOVE;
	}

	bxt_dbg("recv %d: cond %x", fd, cond);

	if (cond & (G_IO_HUP | G_IO_ERR | G_IO_NVAL)) {
		bxt_err("recv %d: IO cond %x errno %d", fd, cond, errno);

		cli->fd_id = 0;
		g_idle_add(close_conn, cli);
		pthread_mutex_unlock(&clients_lock);
		return G_SOURCE_REMOVE;
	}

	r = proc_msg(cli);
	if (r == -1) {
		cli->fd_id = 0;
		g_idle_add(close_conn, cli);
		pthread_mutex_unlock(&clients_lock);
		return G_SOURCE_REMOVE;
	}

	pthread_mutex_unlock(&clients_lock);
	return G_SOURCE_CONTINUE;
}

EXPORT int buxton_close(struct buxton_client *client)
{
	GList *l;

	if (!client) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	pthread_mutex_lock(&clients_lock);
	l = g_list_find(clients, client);
	if (!l) {
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(BUXTON_ERROR_NOT_EXIST);
		return -1;
	}

	free_client(client);

	pthread_mutex_unlock(&clients_lock);
	return 0;
}

EXPORT int buxton_open_full(struct buxton_client **client, bool attach_fd,
		buxton_status_callback callback, void *user_data)
{
	struct buxton_client *cli;
	int r;

	if (!client) {
		bxt_err("Invalid parameter");
		buxton_err_set_errno(BUXTON_ERROR_INVALID_PARAMETER);
		return -1;
	}

	pthread_mutex_lock(&clients_lock);
	cli = calloc(1, sizeof(*cli));
	if (!cli) {
		bxt_err("out of memory");
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(BUXTON_ERROR_OUT_OF_MEMORY);
		return -1;
	}

	cli->fd = -1;
	cli->st_callback = callback;
	cli->st_data = user_data;

	pthread_mutex_init(&cli->lock, NULL);
	pthread_mutex_lock(&cli->lock);
	cli->req_cbs = g_hash_table_new_full(g_direct_hash, g_direct_equal,
			NULL, (GDestroyNotify)free_req);
	if (!cli->req_cbs) {
		pthread_mutex_unlock(&cli->lock);
		free_client(cli);
		bxt_err("out of memory");
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(BUXTON_ERROR_OUT_OF_MEMORY);
		return -1;
	}
	cli->noti_cbs = g_hash_table_new_full(g_str_hash, g_str_equal,
			NULL, (GDestroyNotify)free_noti);
	if (!cli->noti_cbs) {
		pthread_mutex_unlock(&cli->lock);
		free_client(cli);
		bxt_err("out of memory");
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(BUXTON_ERROR_OUT_OF_MEMORY);
		return -1;
	}
	pthread_mutex_unlock(&cli->lock);

	r = connect_server(SOCKPATH, &cli->fd);
	if (r != BUXTON_ERROR_NONE) {
		free_client(cli);
		pthread_mutex_unlock(&clients_lock);
		buxton_err_set_errno(r);
		return -1;
	}

	if (attach_fd)
		cli->fd_id = g_unix_fd_add(cli->fd,
				G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
				recv_cb, cli);
	else
		cli->fd_id = 0;

	clients = g_list_append(clients, cli);
	*client = cli;
	pthread_mutex_unlock(&clients_lock);

	if (callback)
		callback(BUXTON_STATUS_CONNECTED, user_data);

	return 0;
}

EXPORT int buxton_open(struct buxton_client **client,
		buxton_status_callback callback, void *user_data)
{
	return buxton_open_full(client, true, callback, user_data);
}

